#include "acdptu.h"

#include "dptu_c_types.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "dptu_driver.h"

static genom_event ptu_error(const char *error, genom_context self);

/* --- Task cmd --------------------------------------------------------- */

/* --- Activity open ---------------------------------------------------- */

/** Codel dptu_open of activity open.
 *
 * Triggered by dptu_start.
 * Yields to dptu_update.
 * Throws dptu_e_ptu.
 */
genom_event
dptu_open(const char device[64], dptu_baudrate_profile baudrate,
          dptu_ids *ids, genom_context self)
{

  dptu_t *p = (dptu_t *)ids->ptu;
  int br;

  if (p == NULL) {
    ids->ptu = malloc(sizeof(dptu_t));
    p = (dptu_t *)ids->ptu;
  }

  if (baudrate == dptu_BAUDRATE_DEFAULT) {
    br = BAUDRATE_DEFAULT;
  } else { // dptu_BAUDRATE_HIGHEST
    br = BAUDRATE_HIGHEST;
  }

  if (ptuInit(p, device, br) <= 0) {
    printf("ERROR: PTU INIT failed\n");
    return ptu_error("unable to init ptu unit", self);
  }

  /* skeleton sample */ return dptu_update;
}

/** Codel dptu_update_state_poster of activity open.
 *
 * Triggered by dptu_update.
 * Yields to dptu_ether.
 * Throws dptu_e_ptu.
 */
genom_event
dptu_update_state_poster(dptu_ids *ids,
                         const dptu_ptu_state *ptu_state,
                         genom_context self)
{
  dptu_t *p = (dptu_t *)ids->ptu;

  if (p == NULL) {
    return ptu_error("PTU driver not initialised", self);
  }

  dptu_ptu_state_t *state = ptu_state->data(self);
  state->unit = dptu_STEP;
  state->unit_res = dptu_RAD;

  state->tilt_state.speed = p->state.stilt.speed;
  state->tilt_state.base_speed = p->state.stilt.base_speed;
  state->tilt_state.acceleration = p->state.stilt.acc;
  state->tilt_state.position = p->state.stilt.pos;
  state->tilt_state.min_position = p->state.stilt.min_pos;
  state->tilt_state.max_position = p->state.stilt.max_pos;
  state->tilt_state.resolution = p->state.rad_in_step;

  state->pan_state.speed = p->state.span.speed;
  state->pan_state.base_speed = p->state.span.base_speed;
  state->pan_state.acceleration = p->state.span.acc;
  state->pan_state.position = p->state.span.pos;
  state->pan_state.min_position = p->state.span.min_pos;
  state->pan_state.max_position = p->state.span.max_pos;
  state->pan_state.resolution = p->state.rad_in_step;

  ptu_state->write(self);

  /* skeleton sample: insert your code */
  /* skeleton sample */ return dptu_ether;
}

/* --- Activity set_baudrate -------------------------------------------- */

/** Codel dptu_set_baudrate of activity set_baudrate.
 *
 * Triggered by dptu_start.
 * Yields to dptu_ether.
 * Throws dptu_e_ptu.
 */
genom_event
dptu_set_baudrate(const char device[64],
                  dptu_baudrate_profile current_baudrate,
                  dptu_baudrate_profile new_baudrate,
                  dptu_ids *ids, genom_context self)
{
  dptu_t p;
  int br;
  dptu_e_ptu_detail d;

  if (ids->ptu != NULL) {
    return ptu_error("ptu unit connection already open.", self);
  }

  if (current_baudrate == dptu_BAUDRATE_DEFAULT) {
    br = BAUDRATE_DEFAULT;
  } else { // dptu_BAUDRATE_HIGHEST
    br = BAUDRATE_HIGHEST;
  }

  if (ptuInit(&p, device, br) <= 0) {
    return ptu_error("unable to init ptu unit.", self);
  }

  if (new_baudrate == dptu_BAUDRATE_DEFAULT) {
    br = BAUDRATE_DEFAULT;
  } else { // dptu_BAUDRATE_HIGHEST
    br = BAUDRATE_HIGHEST;
  }

  if (ptuChangeBaudrate(p.fd, br) < 0) {
    return ptu_error("unable to change ptu baudrate.", self);
  }

  if (ptuClose(&p) < 0) {
    return ptu_error("unable to close ptu.", self);
  }

  /* skeleton sample: insert your code */
  /* skeleton sample */ return dptu_ether;
}

/* --- Activity goto_position ------------------------------------------- */

/** Codel dptu_goto_position of activity goto_position.
 *
 * Triggered by dptu_start.
 * Yields to dptu_ether.
 * Throws dptu_e_ptu.
 */
genom_event
dptu_goto_position(dptu_ptu_axis axis, double value,
                   dptu_ptu_unit unit, dptu_ids *ids,
                   genom_context self)
{
  dptu_t *p = (dptu_t *)ids->ptu;

  if (p == NULL) {
    return ptu_error("PTU driver not initialised", self);
  }

  const char *cmd;
  if (axis == dptu_PAN) {
    cmd = CMD_ABS_PAN_POS;
  } else { // dptu_TILT
    cmd = CMD_ABS_TILT_POS;
  }

  long svalue;
  if (unit == dptu_STEP) {
    svalue = value;
  } else if (unit == dptu_DEG) {
    svalue = PT_DEG_TO_STEP(p, value);
  } else { // dptu_RAD
    svalue = PT_RAD_TO_STEP(p, value);
  }

  if (ptuCmdSet(p, cmd, svalue) < 0) {
    return ptu_error("Unable to send ptu set command", self);
  }
  /* skeleton sample */ return dptu_ether;
}

/* --- Activity set_speed_profile --------------------------------------- */

/** Codel dptu_set_speed_profile of activity set_speed_profile.
 *
 * Triggered by dptu_start.
 * Yields to dptu_update.
 * Throws dptu_e_ptu.
 */
genom_event
dptu_set_speed_profile(dptu_ptu_axis axis, double speed,
                       double base_speed, double acceleration,
                       dptu_ptu_unit unit, dptu_ids *ids,
                       genom_context self)
{
  dptu_t *p = (dptu_t *)ids->ptu;
  int error = 0;

  if (p == NULL) {
    return ptu_error("PTU not initialised", self);
  }

  long sspeed, sbase_speed, sacceleration;
  if (unit == dptu_STEP) {
    sspeed = speed;
    sbase_speed = base_speed;
    sacceleration = acceleration;
  } else if (unit == dptu_DEG) {
    sspeed = PT_DEG_TO_STEP(p, speed);
    sbase_speed = PT_DEG_TO_STEP(p, base_speed);
    sacceleration = PT_DEG_TO_STEP(p, acceleration);
  } else { // dptu_RAD
    sspeed = PT_RAD_TO_STEP(p, speed);
    sbase_speed = PT_RAD_TO_STEP(p, base_speed);
    sacceleration = PT_RAD_TO_STEP(p, acceleration);
  }

  if (axis == dptu_PAN) {
    error = ptuCmdSet(p, CMD_ABS_PAN_SPEED, sspeed);
    if (error < 0) return ptu_error("Unable to send ptu set command", self);
    error = ptuCmdSet(p, CMD_BASE_PAN_SPEED, sbase_speed);
    if (error < 0) return ptu_error("Unable to send ptu set command", self);
    error = ptuCmdSet(p, CMD_ABS_PAN_ACC, sacceleration);
    if (error < 0) return ptu_error("Unable to send ptu set command", self);
  } else { // dptu_TILT
    error = ptuCmdSet(p, CMD_ABS_TILT_SPEED, sspeed);
    if (error < 0) return ptu_error("Unable to send ptu set command", self);
    error = ptuCmdSet(p, CMD_BASE_TILT_SPEED, sbase_speed);
    if (error < 0) return ptu_error("Unable to send ptu set command", self);
    error = ptuCmdSet(p, CMD_ABS_TILT_ACC, sacceleration);
    if (error < 0) return ptu_error("Unable to send ptu set command", self);
  }

  if (ptuUpdateState(p) < 0) {
    return ptu_error("Unable to update ptu state.", self);
  }

  /* skeleton sample */ return dptu_update;
}

/** Codel dptu_update_state_poster of activity set_speed_profile.
 *
 * Triggered by dptu_update.
 * Yields to dptu_ether.
 * Throws dptu_e_ptu.
 */
/* already defined in service open */


static genom_event ptu_error(const char *error, genom_context self) {
  dptu_e_ptu_detail d;
  strcpy(d.what, error);
  return dptu_e_ptu(&d, self);
}
